
/*
  Kjetil Matheussen, 2005/2006. Based on:


    Copyright (C) 2001 Paul Davis
    Copyright (C) 2003 Jack O'Quin
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    * 2002/08/23 - modify for libsndfile 1.0.0 <andy@alsaplayer.org>
    * 2003/05/26 - use ringbuffers - joq
    
    $Id: capture_client.c,v 1.11 2004/03/27 22:08:55 joq Exp $
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sndfile.h>
#include <pthread.h>
#include <getopt.h>

#include <jack/jack.h>
#include <jack/ringbuffer.h>

#define JC_MAX(a,b) (((a)>(b))?(a):(b))
#define JC_MIN(a,b) (((a)<(b))?(a):(b))

typedef struct _thread_info {
    pthread_t thread_id;
    SNDFILE *sf;
    jack_nframes_t rb_size;
    jack_client_t *client;
    unsigned int channels;
    int bitdepth;
    char *path;
    int leading_zeros;
    volatile int can_capture;
    volatile int can_process;
    volatile int status;
} jack_thread_info_t;

jack_thread_info_t *global_thread_info_t;

/* JACK data */
unsigned int nports;
jack_port_t **ports;
jack_default_audio_sample_t **in;
jack_nframes_t nframes;
const size_t sample_size = sizeof(jack_default_audio_sample_t);
long overruns = 0;


/* Synchronization between process thread and disk thread. */
#define DEFAULT_RB_SIZE (16384*32)		/* ringbuffer size in frames */
jack_ringbuffer_t *rb;
pthread_mutex_t disk_thread_lock = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t  data_ready = PTHREAD_COND_INITIALIZER;

/* Jack connecting thread. */
pthread_mutex_t connect_thread_lock = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t  connect_ready = PTHREAD_COND_INITIALIZER;
pthread_t connect_thread={0} ;


static volatile int is_running=1;


void *disk_thread (void *arg){
	jack_thread_info_t *info = (jack_thread_info_t *) arg;
	jack_nframes_t samples_per_frame = info->channels;
	size_t bytes_per_frame = samples_per_frame * sample_size;
	void *framebuf = malloc (bytes_per_frame);

	pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
	pthread_mutex_lock (&disk_thread_lock);

	info->status = 0;

	for(;;){
	  
	  /* Write the data one frame at a time.  This is
	   * inefficient, but makes things simpler. */
	  while (info->can_capture &&
		 (jack_ringbuffer_read_space (rb) >= bytes_per_frame)) {
	    
	    jack_ringbuffer_read (rb, framebuf, bytes_per_frame);
	    
	    if (sf_writef_float (info->sf, framebuf, 1) != 1) {
	      char errstr[256];
	      sf_error_str (0, errstr, sizeof (errstr) - 1);
	      fprintf (stderr,
		       "cannot write sndfile (%s)\n",
		       errstr);
	      info->status = EIO; /* write failed */
	      goto done;
	    }
	  }
	  
	  if (!is_running) {
	    goto done;
	  }

	  /* wait until process() signals more data */
	  pthread_cond_wait (&data_ready, &disk_thread_lock);
	}


 done:
	pthread_mutex_unlock (&disk_thread_lock);
	free (framebuf);
	printf ("disk thread finished\n");
	return 0;
}

	
int process (jack_nframes_t nframes, void *arg){
	int chn;
	size_t i;
	jack_thread_info_t *info = (jack_thread_info_t *) arg;

	/* Do nothing until we're ready to begin. */
	if ((!info->can_process) || (!info->can_capture))
		return 0;

	for (chn = 0; chn < nports; chn++)
		in[chn] = jack_port_get_buffer (ports[chn], nframes);

	/* Sndfile requires interleaved data.  It is simpler here to
	 * just queue interleaved samples to a single ringbuffer. */
	for (i = 0; i < nframes; i++) {
		for (chn = 0; chn < nports; chn++) {
			if (jack_ringbuffer_write (rb, (void *) (in[chn]+i),
					      sample_size)
			    < sample_size)
				overruns++;
		}
	}

	/* Tell the disk thread there is work to do.  If it is already
	 * running, the lock will not be available.  We can't wait
	 * here in the process() thread, but we don't need to signal
	 * in that case, because the disk thread will read all the
	 * data queued before waiting again. */
	if (pthread_mutex_trylock (&disk_thread_lock) == 0) {
	    pthread_cond_signal (&data_ready);
	    pthread_mutex_unlock (&disk_thread_lock);
	}

	return 0;
}


void
setup_disk_thread (jack_thread_info_t *info)
{
	SF_INFO sf_info;
	int short_mask;
	
	sf_info.samplerate = jack_get_sample_rate (info->client);
	sf_info.channels = info->channels;

	switch (info->bitdepth) {
		case 8: short_mask = SF_FORMAT_PCM_U8;
		  	break;
		case 16: short_mask = SF_FORMAT_PCM_16;
			 break;
		case 24: short_mask = SF_FORMAT_PCM_24;
			 break;
		case 32: short_mask = SF_FORMAT_PCM_32;
			 break;
		default: short_mask = SF_FORMAT_FLOAT;
			 break;
	}		 
	sf_info.format = SF_FORMAT_WAV|short_mask;

	if ((info->sf = sf_open (info->path, SFM_WRITE, &sf_info)) == NULL) {
		char errstr[256];
		sf_error_str (0, errstr, sizeof (errstr) - 1);
		fprintf (stderr, "cannot open sndfile \"%s\" for output (%s)\n", info->path, errstr);
		jack_client_close (info->client);
		exit (1);
	}

	info->can_capture = 0;

	pthread_create (&info->thread_id, NULL, disk_thread, info);
}

void start_disk_thread (jack_thread_info_t *info){
  info->can_capture = 1;
}

void stop_disk_thread(jack_thread_info_t *info){
 
 /* Wake up disk thread. (no trylock) (isrunning==0)*/
  pthread_cond_signal (&data_ready);

  pthread_join (info->thread_id, NULL);
  sf_close (info->sf);
  if (overruns > 0) {
    fprintf (stderr,
	     "jackrec failed with %ld overruns.\n", overruns);
    fprintf (stderr, " try a bigger buffer than -B %"
	     PRIu32 ".\n", info->rb_size);
    info->status = EPIPE;
  }
  if (info->status) {
    unlink (info->path);
  }
}

static volatile int dontreorderyet=1;


static int findnumconnections(const char **connections){
  int ret=0;
  while(connections && connections[ret]!=NULL)
    ret++;
  return ret;
}

static int compaire(const void *a, const void *b){
  return strcmp((const char*)a,(const char*)b);
}

void connect_ports_in_thread(jack_thread_info_t *info,jack_client_t *client){

  // First check if anything needs to be done, return if not.
  //printf("connect_ports\n");
  {
    int ch=0;
    const char **outportnames;

    outportnames=jack_get_ports(client,NULL,NULL,JackPortIsPhysical|JackPortIsInput);

    ch=0;
    while(outportnames[ch]!=NULL && ch<info->channels){
      int lokke = 0;
      //fprintf(stderr,"ai %x\n",jack_port_by_name(info->client,outportnames[ch]));
      const char **connections1 = jack_port_get_all_connections(client,jack_port_by_name(client,outportnames[ch]));
      const char **connections2 = jack_port_get_all_connections(client,ports[ch]);
      int memb1=findnumconnections(connections1);
      int memb2=findnumconnections(connections2);

      if(memb1!=memb2)
	goto somethingneedstobedone;
      //      printf("same size\n");

      qsort(connections1,memb1,sizeof(char*),compaire);
      qsort(connections2,memb1,sizeof(char*),compaire);

      for(lokke=0;lokke<memb1;lokke++){
	//printf("connect_ports \"%s\" \"%s\" \n",connections1[lokke],connections2[lokke]);
	if(strcmp(connections1[lokke],connections2[lokke]))
	  goto somethingneedstobedone;
      }

      ch++;
    }
  }
  return;


 somethingneedstobedone:

  printf("Reconnecting ports.\n");

  // disconnect
  //outportnames=jack_get_ports(client,"jack_capture*",NULL,0);
  {
    int ch;
    for(ch=0;ch<info->channels;ch++){
      int lokke = 0;
      //fprintf(stderr,"ch: %d %s %x %x\n",ch,outportnames[ch],ports[ch],jack_port_by_name(info->client,outportnames[ch]));
      const char **connections = jack_port_get_all_connections(client,ports[ch]);
      //fprintf(stderr,"ch2: %d %x\n",ch,connections);
      while(connections && connections[lokke] != NULL){
	//printf("hm %s %s\n",connections[lokke],jack_port_name(ports[ch]));
	jack_disconnect(client,connections[lokke],jack_port_name(ports[ch]));
	lokke++;
      }
    }
  }

  // connect
  {
    const char **outportnames=jack_get_ports(client,NULL,NULL,JackPortIsPhysical|JackPortIsInput);
    int ch=0;
    while(outportnames[ch]!=NULL && ch<info->channels){
      int lokke = 0,err;
      //fprintf(stderr,"ai %x\n",jack_port_by_name(info->client,outportnames[ch]));
      const char **connections = jack_port_get_all_connections(client,jack_port_by_name(client,outportnames[ch]));
      //fprintf(stderr,"%s %x..\n",outportnames[ch],connections);
      while(connections && connections[lokke] != NULL){
	//fprintf(stderr,"%s<<<\n",connections[lokke]);
	if ((err=jack_connect (client, connections[lokke], jack_port_name (ports[ch])))) {
	  fprintf (stderr, "cannot connect input port %s to %s, errorcode %s\n", jack_port_name (ports[ch]), connections[lokke],strerror(err));
	  //jack_client_close (client);
	  //exit (1);
	} 
	
	lokke++;
      }
      ch++;
    }
  }

}


void* connection_thread(void *arg){
  jack_thread_info_t *info = (jack_thread_info_t *) arg;
  jack_client_t *client;

  pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
  pthread_mutex_lock (&connect_thread_lock);

  if ((client = jack_client_new ("jack_connect_capture")) == 0) {
    fprintf (stderr, "jack server not running?\n");
    exit (1);
  }

  connect_ports_in_thread(info,client);
  dontreorderyet=0;

  info->can_process = 1;		/* process() can start, now */

  while(1){
    pthread_cond_wait (&connect_ready, &connect_thread_lock);
    if (!is_running) {
      goto done;
    }
    connect_ports_in_thread(info,client);
  }

 done:
  jack_client_close (client);
  pthread_mutex_unlock (&connect_thread_lock);
  printf ("connection thread finished\n");
  return NULL;
}


void wake_up_connection_thread(void){
  // Don't want trylock here.
  pthread_cond_signal (&connect_ready);
}


void setup_ports (jack_thread_info_t *info){
	unsigned int i;
	size_t in_size;

	/* Allocate data structures that depend on the number of ports. */
	nports = info->channels;
	ports = (jack_port_t **) malloc (sizeof (jack_port_t *) * nports);
	in_size =  nports * sizeof (jack_default_audio_sample_t *);
	in = (jack_default_audio_sample_t **) malloc (in_size);
	rb = jack_ringbuffer_create (nports * sample_size * info->rb_size);

	/* When JACK is running realtime, jack_activate() will have
	 * called mlockall() to lock our pages into memory.  But, we
	 * still need to touch any newly allocated pages before
	 * process() starts using them.  Otherwise, a page fault could
	 * create a delay that would force JACK to shut us down. */
	memset(in, 0, in_size);
	memset(rb->buf, 0, rb->size);

	for (i = 0; i < nports; i++) {
		char name[500];

		sprintf (name, "input%d", i+1);

		if ((ports[i] = jack_port_register (info->client, name, JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0)) == 0) {
			fprintf (stderr, "cannot register input port \"%s\"!\n", name);
			jack_client_close (info->client);
			exit (1);
		}
	}

	pthread_create (&connect_thread, NULL, connection_thread, info);       
}


int graphordercallback(void *arg){
  //jack_thread_info_t *info = (jack_thread_info_t *) arg;

  if(dontreorderyet==1)
    return 0;

  //dontreorderyet=1;{
  //fprintf(stderr,"jack graphorder!\n");
  //printf("(jack graph reorder)\n");
  wake_up_connection_thread();
  //}dontreorderyet=0;
  return 0;
}


void do_exit(jack_thread_info_t *info,int close_jack){
  is_running=0;

  stop_disk_thread(info);

  if(close_jack)
    jack_client_close(info->client);

  jack_ringbuffer_free (rb);

  wake_up_connection_thread();  //Tell connect thread to quit as well.
  pthread_join(connect_thread, NULL); //Wait for it.
}


void jack_shutdown (void *arg){
  fprintf (stderr, "JACK shutdown.\n");
  do_exit((jack_thread_info_t*)arg,0);
  exit (0);
  //abort();
}

int main (int argc, char *argv[]){
  jack_client_t *client;
	jack_thread_info_t thread_info={0};
	int c;
	int longopt_index = 0;
	extern int optind, opterr;
	int show_usage = 0;
	char *optstring = "c:b:B:z:h";
	struct option long_options[] = {
		{ "help", 0, 0, 'h' },
		{ "channels", 1, 0, 'c' },
		{ "bitdepth", 1, 0, 'b' },
		{ "bufsize", 1, 0, 'B' },
		{ "leading-zeros", 1, 0, 'z'},
		{ 0, 0, 0, 0 }
	};


	{
	  thread_info.rb_size = DEFAULT_RB_SIZE;
	  thread_info.channels = 2;
	  opterr = 0;
	  
	  while ((c = getopt_long (argc, argv, optstring, long_options, &longopt_index)) != -1) {
	    switch (c) {
	    case 1:
	      /* getopt signals end of '-' options */
	      break;
	      
	    case 'h':
	      show_usage++;
	      break;
	      //case 'f':
	      //			thread_info.path = optarg;
	      //break;
	    case 'b':
	      thread_info.bitdepth = atoi (optarg);
	      break;
	    case 'c':
	      thread_info.channels=atoi(optarg);
	      break;
	    case 'B':
	      thread_info.rb_size = atoi (optarg);
	      break;
	    case 'z':
	      thread_info.leading_zeros = JC_MIN(8,atoi(optarg));
	      break;
	    default:
	      fprintf (stderr, "error\n");
	      show_usage++;
	      break;
	    }
	  }
	  
	  if(optind!=argc)
	    thread_info.path = argv[optind];
	  //printf("%d %d %s\n",optind,argc,argv[optind]);
	  //	if(optind>argc)
	  //	setup_ports (argc - optind, &argv[optind], &thread_info);
	  
	  if (show_usage) {
	    fprintf (stderr, "usage: jack_capture [ -b bitdepth ] [ -B bufsize ] [-c channels] [-z leading-zeros] [-h] [filename]\n");
	    exit (1);
	  }
	  
	  if ((client = jack_client_new ("jack_capture")) == 0) {
	    fprintf (stderr, "jack server not running?\n");
	    exit (1);
	  }
	  
	  thread_info.client = client;
	  thread_info.can_process = 0;
	  
	  if(thread_info.path==NULL){
	    int try=0;
	    thread_info.path=malloc(5000);
	    for(;;){
	      const char *temp[20]={"jack_capture_%d.wav","jack_capture_%01d.wav","jack_capture_%02d.wav","jack_capture_%03d.wav","jack_capture_%04d.wav","jack_capture_%05d.wav","jack_capture_%06d.wav","jack_capture_%07d.wav","jack_capture_%08d.wav","jack_capture_%09d.wav"};
	      
	      sprintf(thread_info.path,temp[thread_info.leading_zeros+1],++try);
	      if(access(thread_info.path,F_OK)) break;
	    }
	    
	  }
	  

	  setup_disk_thread (&thread_info);
	  
	  {
	    jack_set_process_callback (client, process, &thread_info);
	    jack_on_shutdown (client, jack_shutdown, &thread_info);
	    
	    
	    jack_set_graph_order_callback(client,graphordercallback,&thread_info);
	  
	    if (jack_activate (client)) {
	      fprintf (stderr, "cannot activate client");
	    }
	  }
	  
	  setup_ports (&thread_info);
	  start_disk_thread (&thread_info);
	}

	global_thread_info_t=&thread_info;

	printf("Recording to \"%s\". Press <Ctrl-C> to stop it.\n",thread_info.path);
	{
	  char gakk[64];
	  fgets(gakk,49,stdin);
	}

	{
	  do_exit(&thread_info,1);
	}

	return 0;
}


